from typing import Annotated
from httpx import Client
from collections import defaultdict
from dateutil import parser
from datetime import datetime, timezone, timedelta

from admyral.action import action, ArgumentMetadata

from admyral.context import ctx
from admyral.typings import JsonValue


def get_retool_client(domain: str, api_key: str) -> Client:
    # Auth: https://docs.retool.com/org-users/guides/retool-api/authentication#tag/Organization/paths/~1usage~1organizations/get
    return Client(
        base_url=f"https://{domain}/api/v2",
        headers={
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="List Inactive Retool Users",
    display_namespace="Retool",
    description="This method lists Retool users who have not logged in for a specified number of days.",
    secrets_placeholders=["RETOOL_SECRET"],
)
def list_retool_inactive_users(
    inactivity_threshold_in_days: Annotated[
        int,
        ArgumentMetadata(
            display_name="Inactivity Threshold in Days",
            description="The number of days since the last login after which a user is considered inactive.",
        ),
    ] = 60,
) -> list[dict[str, JsonValue]]:
    # Returns a list of users. The API token must have the "Users > Read" scope.
    # https://docs.retool.com/reference/api/#tag/Users/paths/~1users~1%7BuserId%7D/patch

    secret = ctx.get().secrets.get("RETOOL_SECRET")
    api_key = secret["api_key"]
    domain = secret["domain"]

    with get_retool_client(domain, api_key) as client:
        # TODO: handle pageination
        response = client.get("/users")
        response.raise_for_status()
        response_json = response.json()

        if not response_json["success"]:
            raise RuntimeError(
                f"Failed to list Retool users: {response_json['message']}"
            )

        utc_time_now = datetime.now(timezone.utc) - timedelta(
            days=inactivity_threshold_in_days
        )

        return [
            user
            for user in response_json["data"]
            if parser.isoparse(user["last_active"]) <= utc_time_now
        ]


@action(
    display_name="List Groups per User",
    display_namespace="Retool",
    description="List all groups a user is a member of as well as the last time a user was active in Retool.",
    secrets_placeholders=["RETOOL_SECRET"],
)
def list_groups_per_user() -> dict[str, JsonValue]:
    # Get all permission groups for an organization or space. The API token must have the "Groups > Read" scope.
    # https://docs.retool.com/reference/api/v2/#tag/Groups/paths/~1groups~1%7BgroupId%7D/put

    secret = ctx.get().secrets.get("RETOOL_SECRET")
    api_key = secret["api_key"]
    domain = secret["domain"]

    with get_retool_client(domain, api_key) as client:
        # Fetch all users to get the last active date
        response = client.get("/users")
        response.raise_for_status()
        response_json = response.json()

        if not response_json["success"]:
            raise RuntimeError(
                f"Failed to list Retool users: {response_json['message']}"
            )

        groups_per_user = {
            user["email"]: {
                "groups": defaultdict(list),
                "last_active": user["last_active"],
            }
            for user in response_json["data"]
        }

        # Fetch all groups
        response = client.get("/groups")
        response.raise_for_status()
        response_json = response.json()

        if not response_json["success"]:
            raise RuntimeError(
                f"Failed to list Retool groups: {response_json['message']}"
            )

        for group in response_json["data"]:
            for member in group["member"]:
                groups_per_user[member]["groups"].append(group["name"])

        return groups_per_user



from admyral.utils.singleton import Singleton
from admyral.db.store_interface import StoreInterface
from admyral.secret.secrets_manager import SecretsManager


class SharedWorkerState(metaclass=Singleton):
    _store: StoreInterface = None
    _secrets_manager: SecretsManager = None

    @classmethod
    def init(cls, store: StoreInterface, secrets_manager: SecretsManager) -> None:
        cls._store = store
        cls._secrets_manager = secrets_manager

    @classmethod
    def get_store(cls) -> StoreInterface:
        if not cls._store:
            raise RuntimeError("SharedWorkerState not initialized.")
        return cls._store

    @classmethod
    def get_secrets_manager(cls) -> SecretsManager:
        if not cls._secrets_manager:
            raise RuntimeError("SharedWorkerState not initialized.")
        return cls._secrets_manager



from temporalio import activity
import time


from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.logger import get_logger
from admyral.utils.time import utc_now


logger = get_logger(__name__)


@activity.defn
async def mark_workflow_as_completed(
    run_id: str,
) -> str:
    start = time.monotonic_ns()
    await SharedWorkerState.get_store().mark_workflow_run_as_completed(
        run_id=run_id,
        completed_at=utc_now(),
    )
    end = time.monotonic_ns()
    logger.info(
        f"Marked workflow run {run_id} as completed in {(end - start) / 1_000_000}ms"
    )



from typing import Literal
from pydantic import BaseModel
from enum import Enum

from admyral.models.action import ActionMetadata


########################################################
# Editor Actions
########################################################


class ActionNamespace(BaseModel):
    namespace: str
    actions: list[ActionMetadata]


class EditorActions(BaseModel):
    control_flow_actions: list[ActionMetadata]
    namespaces: list[ActionNamespace]


########################################################
# Editor Workflow Graph
########################################################


class EditorWorkflowBaseNode(BaseModel):
    id: str
    type: str
    """ Node Type """
    action_type: str
    """ Action Type """


class EditorWebhookTrigger(BaseModel):
    webhook_id: str | None
    webhook_secret: str | None
    default_args: list[tuple[str, str]]


class EditorScheduleType(str, Enum):
    CRON = "Cron"
    INTERVAL_SECONDS = "Interval Seconds"
    INTERVAL_MINUTES = "Interval Minutes"
    INTERVAL_HOURS = "Interval Hours"
    INTERVAL_DAYS = "Interval Days"


class EditorScheduleTrigger(BaseModel):
    schedule_type: EditorScheduleType
    value: str
    default_args: list[tuple[str, str]]


class EditorWorkflowStartNode(EditorWorkflowBaseNode):
    type: Literal["start"] = "start"
    action_type: Literal["start"] = "start"

    webhook: EditorWebhookTrigger | None
    schedules: list[EditorScheduleTrigger]


class EditorWorkflowActionNode(EditorWorkflowBaseNode):
    type: Literal["action"] = "action"
    result_name: str | None
    secrets_mapping: dict[str, str]
    args: dict[str, str]


class EditorWorkflowIfNode(EditorWorkflowBaseNode):
    type: Literal["if_condition"] = "if_condition"
    action_type: Literal["if_condition"] = "if_condition"
    condition: str


type EditorWorkflowNode = (
    EditorWorkflowStartNode | EditorWorkflowActionNode | EditorWorkflowIfNode
)


class EditorWorkflowEdgeType(str, Enum):
    DEFAULT = "default"
    TRUE = "true"
    FALSE = "false"


class EditorWorkflowEdge(BaseModel):
    source: str
    target: str
    type: EditorWorkflowEdgeType


class EditorWorkflowGraph(BaseModel):
    workflow_id: str
    workflow_name: str
    description: str | None
    is_active: bool
    nodes: list[EditorWorkflowNode]
    edges: list[EditorWorkflowEdge]



class AdmyralError(Exception):
    """Base for all Admyral exceptions."""

    @property
    def cause(self) -> BaseException | None:
        """Cause of the exception.

        This is the same as ``Exception.__cause__``.
        """
        return self.__cause__


class AdmyralFailureError(AdmyralError):
    """Base for runtime failures during workflow/action execution."""

    def __init__(
        self,
        message: str,
    ) -> None:
        super().__init__(message)
        self._message = message

    @property
    def message(self) -> str:
        """Message."""
        return self._message


class NonRetryableActionError(AdmyralFailureError):
    """Raised when an action is not retryable."""

    def __init__(
        self,
        message: str,
    ) -> None:
        super().__init__(message)


class RetryableActionError(AdmyralFailureError):
    """Raised when an action is retryable."""

    def __init__(
        self,
        message: str,
    ) -> None:
        super().__init__(message)



from admyral.actions.integrations.cdr.ms_defender_for_cloud import (
    list_ms_defender_for_cloud_alerts,
)
from admyral.actions.integrations.cdr.wiz import list_wiz_alerts

__all__ = ["list_ms_defender_for_cloud_alerts", "list_wiz_alerts"]



from typing import Any


def _is_json_value(obj: Any) -> bool:
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return True
    elif isinstance(obj, list):
        return all(_is_json_value(item) for item in obj)
    elif isinstance(obj, dict):
        return all(
            isinstance(key, str) and _is_json_value(value) for key, value in obj.items()
        )
    else:
        return False


def is_allowed_return_type(obj: Any) -> bool:
    return _is_json_value(obj)


def throw_if_not_allowed_return_type(obj: Any) -> None:
    if not is_allowed_return_type(obj):
        raise ValueError(
            f"Object of type {type(obj)} is not allowed in the return statement. The type must be JSON."
        )



import os
from admyral.logger import setup_root_logger, get_logger
import tomllib

setup_root_logger()

logger = get_logger(__name__)


# Set the version number
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

if os.path.exists(os.path.join(ROOT_DIR, "VERSION")):
    with open(os.path.join(ROOT_DIR, "VERSION")) as version_file:
        __version__ = version_file.read().strip()
elif os.path.exists(os.path.join(ROOT_DIR, "..", "pyproject.toml")):
    with open(os.path.join(ROOT_DIR, "..", "pyproject.toml"), "rb") as project_file:
        __version__ = tomllib.load(project_file)["tool"]["poetry"]["version"]
else:
    logger.warn("Could not determine admyral version number.")
    __version__ = None



from admyral.server.endpoints.action_endpoints import router as action_router
from admyral.server.endpoints.secret_endpoints import router as secret_router
from admyral.server.endpoints.webhook_endpoints import router as webhook_router
from admyral.server.endpoints.workflow_endpoints import router as workflow_router
from admyral.server.endpoints.editor_endpoints import router as editor_router
from admyral.server.endpoints.workflow_run_endpoints import (
    router as workflow_run_router,
)

__all__ = [
    "action_router",
    "secret_router",
    "webhook_router",
    "workflow_router",
    "editor_router",
    "workflow_run_router",
]



import click
import os
from enum import Enum
import logging
from dotenv import load_dotenv
from pydantic import BaseModel

load_dotenv()


class LoggingLevel(Enum):
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


ENV_ADMYRAL_APP_DIR = "ADMYRAL_APP_DIR"
ENV_ADMYRAL_LOGGING_LEVEL = "ADMYRAL_LOGGING_LEVEL"
ENV_ADMYRAL_DISABLE_NSJAIL = "ADMYRAL_DISABLE_NSJAIL"
ENV_ADMYRAL_FLOCK_PATH = "ADMYRAL_FLOCK_PATH"
ENV_ADMYRAL_PIP_CACHE_DIRECTORY = "ADMYRAL_PIP_CACHE_DIRECTORY"
ENV_ADMYRAL_PIP_LOCK_CACHE_DIRECTORY = "ADMYRAL_PIP_LOCK_CACHE_DIRECTORY"
ENV_ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE = "ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE"


APP_NAME = "Admyral"

# Name of the configuration file
GLOBAL_CONFIG_FILE = "config.yaml"

PROJECT_DIRECTORY = ".admyral"

ADMYRAL_DEFAULT_STORAGE_DIRECTORY_NAME = "admyral_default_storage"

LOGGING_LEVEL = os.getenv(ENV_ADMYRAL_LOGGING_LEVEL, "INFO").upper()


def get_global_project_directory() -> str:
    """
    # TODO: adapt descr.
    Returns the global config folder for Admyral. The default behavior is to return whatever is most appropriate for the operating system.
    """
    return os.environ.get(ENV_ADMYRAL_APP_DIR, click.get_app_dir(APP_NAME))


def get_local_storage_path() -> str:
    """
    TODO: add descr.
    """
    return os.path.join(
        get_global_project_directory(), ADMYRAL_DEFAULT_STORAGE_DIRECTORY_NAME
    )


def get_local_postgres_volume() -> str:
    return os.path.join(get_local_storage_path(), "postgres")


# Constants for custom Python execution
ADMYRAL_DISABLE_NSJAIL = (
    os.getenv(ENV_ADMYRAL_DISABLE_NSJAIL, "false").lower() == "true"
)
ADMYRAL_PIP_CACHE_DIRECTORY = os.getenv(
    ENV_ADMYRAL_PIP_CACHE_DIRECTORY,
    os.path.join(get_local_storage_path(), "cache", "pip"),
)
ADMYRAL_PIP_LOCK_CACHE_DIRECTORY = os.getenv(
    ENV_ADMYRAL_PIP_LOCK_CACHE_DIRECTORY,
    os.path.join(get_local_storage_path(), "cache", "pip-lock"),
)
ADMYRAL_PIP_LOCKFILE_CACHE_TTL_IN_SECONDS = 3 * 24 * 60 * 60  # 3 days
ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE = (
    os.getenv(ENV_ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE, "true").lower() == "true"
)


# DATABASE AND SECRETS MANAGER


class DatabaseType(str, Enum):
    """
    Enum class for the supported database types.
    """

    SQLITE = "sqlite"
    POSTGRES = "postgres"


class SecretsManagerType(str, Enum):
    """
    Enum class for the supported secrets manager types.
    """

    SQL = "sql"
    # AWS_SECRETS_MANAGER = "aws_secrets_manager"


SQLITE_DATABASE_NAME = "admyral.db"

ENV_ADMYRAL_DATABASE_TYPE = "ADMYRAL_DATABASE_TYPE"
ENV_ADMYRAL_DATABASE_URL = "ADMYRAL_DATABASE_URL"
ENV_ADMYRAL_SECRETS_MANAGER_TYPE = "ADMYRAL_SECRETS_MANAGER"


ADMYRAL_DATABASE_TYPE = DatabaseType(
    os.getenv(ENV_ADMYRAL_DATABASE_TYPE, DatabaseType.SQLITE)
)
ADMYRAL_DATABASE_URL = os.getenv(
    ENV_ADMYRAL_DATABASE_URL,
    f"sqlite+aiosqlite:///{get_local_storage_path()}/{SQLITE_DATABASE_NAME}",
)
ADMYRAL_SECRETS_MANAGER_TYPE = SecretsManagerType(
    os.getenv(ENV_ADMYRAL_SECRETS_MANAGER_TYPE, SecretsManagerType.SQL)
)


ENV_TEMPORAL_HOST = "ADMYRAL_TEMPORAL_HOST"

ADMYRAL_TEMPORAL_HOST = os.getenv(ENV_TEMPORAL_HOST, "localhost:7233")


# TODO: read from and persist to disk
class GlobalConfig(BaseModel):
    """
    The global configuration for Admyral.
    """

    user_id: str = "38815447-e272-4299-94c0-29a2d30435f9"
    storage_directory: str = get_local_storage_path()
    database_type: DatabaseType = ADMYRAL_DATABASE_TYPE
    database_url: str = ADMYRAL_DATABASE_URL
    temporal_host: str = ADMYRAL_TEMPORAL_HOST
    secrets_manager_type: SecretsManagerType = ADMYRAL_SECRETS_MANAGER_TYPE

    pip_lockfile_cache_cleanup_interval: int = 60 * 60 * 24  # 1 day


ADMYRAL_DAEMON_PID_FILE = "admyral.pid"
ADMYRAL_DAEMON_LOG_FILE = "admyral.log"

TEMPORAL_DAEMON_PID_FILE = "temporal.pid"
TEMPORAL_DAEMON_LOG_FILE = "temporal.log"


def get_admyral_daemon_directory() -> str:
    return os.path.join(get_global_project_directory(), "Admyral", "daemon")


def get_admyral_daemon_pid_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), ADMYRAL_DAEMON_PID_FILE)


def get_temporal_daemon_pid_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), TEMPORAL_DAEMON_PID_FILE)


def get_admyral_daemon_log_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), ADMYRAL_DAEMON_LOG_FILE)


def get_temporal_daemon_log_file() -> str:
    return os.path.join(get_admyral_daemon_directory(), TEMPORAL_DAEMON_LOG_FILE)


ENV_ADMYRAL_WEBHOOK_SIGNING_SECRET = "ADMYRAL_WEBHOOK_SIGNING_SECRET"
WEBHOOK_SIGNING_SECRET = bytes.fromhex(
    os.getenv(
        ENV_ADMYRAL_WEBHOOK_SIGNING_SECRET,
        "ebedd9f0c10b01acb9fd097cdf34a2b38bb554f2c7f68f0d9f534eff5c6ef5d9",
    )
)

ENV_ADMYRAL_SECRETS_ENCRYPTION_KEY = "ADMYRAL_SECRETS_ENCRYPTION_KEY"
SECRETS_ENCRYPTION_KEY = bytes.fromhex(
    os.getenv(
        ENV_ADMYRAL_SECRETS_ENCRYPTION_KEY,
        "834f01cf391c972f1e6def3d7f315f8194bb10048e5cf282aa4cba63b239d8fb",
    )
)


API_V1_STR = "/api/v1"



from fastapi import APIRouter, status, BackgroundTasks, Body
from typing import Optional, Annotated, Callable, Coroutine
from uuid import uuid4
from pydantic import BaseModel

from admyral.utils.collections import is_not_empty
from admyral.server.deps import get_admyral_store, get_workers_client
from admyral.models import (
    Workflow,
    WorkflowTriggerType,
    WorkflowPushRequest,
    WorkflowPushResponse,
    WorkflowSchedule,
    WorkflowTriggerResponse,
    WorkflowMetadata,
)
from admyral.logger import get_logger
from admyral.typings import JsonValue


logger = get_logger(__name__)


MANUAL_TRIGGER_SOURCE_NAME = "manual"
SPACE = " "


class WorkflowBody(BaseModel):
    workflow_name: str | None = None
    workflow_id: str | None = None


async def _fetch_workflow(
    workflow_name: str | None, workflow_id: str | None
) -> Workflow:
    if not workflow_name and not workflow_id:
        raise ValueError("Either workflow_name or workflow_id must be provided.")

    if workflow_id:
        workflow = await get_admyral_store().get_workflow_by_id(workflow_id)
        if not workflow:
            raise ValueError(f"Workflow with id {workflow_id} not found.")
    else:
        workflow = await get_admyral_store().get_workflow_by_name(workflow_name)
        if not workflow:
            raise ValueError(f"Workflow with name {workflow_name} not found.")

    return workflow


async def push_workflow_impl(
    workflow_name: str, workflow_id: str | None, request: WorkflowPushRequest
) -> WorkflowPushResponse:
    """
    Push a workflow to the store. If the workflow for the provided workflow id already exists,
    it will be overwritten.

    Args:
        workflow_name: The workflow name.
        workflow: The workflow object.
    """
    admyral_store = get_admyral_store()
    workers_client = get_workers_client()

    if SPACE in workflow_name:
        raise ValueError("Workflow name must not contain spaces.")

    if workflow_id:
        existing_workflow = await admyral_store.get_workflow_by_id(workflow_id)
    else:
        existing_workflow = await admyral_store.get_workflow_by_name(workflow_name)

    # fetch workflow webhook and schedules if they exist
    existing_webhook = None
    existing_schedules = []
    if existing_workflow:
        existing_webhook = await admyral_store.get_webhook_for_workflow(
            existing_workflow.workflow_id
        )
        existing_schedules = await admyral_store.list_schedules_for_workflow(
            existing_workflow.workflow_id
        )

    if existing_workflow:
        workflow_id = existing_workflow.workflow_id
    elif not workflow_id:
        workflow_id = str(uuid4())

    workflow = Workflow(
        workflow_id=workflow_id,
        workflow_name=workflow_name,
        workflow_dag=request.workflow_dag,
        is_active=request.activate,
    )

    await admyral_store.store_workflow(workflow)

    # Handle schedule update

    # Delete all existing schedules
    for schedule in existing_schedules:
        if existing_workflow and existing_workflow.is_active:
            # the schedule only exists if we previously had an active workflow
            await workers_client.delete_schedule(schedule.schedule_id)
        await admyral_store.delete_schedule(schedule.schedule_id)

    new_schedules = list(
        filter(
            lambda trigger: trigger.type == WorkflowTriggerType.SCHEDULE,
            request.workflow_dag.start.triggers,
        )
    )
    if is_not_empty(new_schedules):
        for schedule in new_schedules:
            new_workflow_schedule = WorkflowSchedule(
                schedule_id=str(uuid4()),
                workflow_id=workflow.workflow_id,
                cron=schedule.cron,
                interval_seconds=schedule.interval_seconds,
                interval_minutes=schedule.interval_minutes,
                interval_hours=schedule.interval_hours,
                interval_days=schedule.interval_days,
                default_args=schedule.default_args_dict,
            )
            await admyral_store.store_schedule(new_workflow_schedule)

            if request.activate:
                await workers_client.schedule_workflow(workflow, new_workflow_schedule)

    # Handle webhook update
    filtered_webhooks = list(
        filter(
            lambda trigger: trigger.type == WorkflowTriggerType.WEBHOOK,
            request.workflow_dag.start.triggers,
        )
    )
    if len(filtered_webhooks) > 1:
        raise ValueError("Multiple webhooks found for the same workflow.")

    webhook_response = None
    new_webhook = filtered_webhooks[0] if is_not_empty(filtered_webhooks) else None
    if new_webhook:
        if not existing_webhook:
            # Case: we didn't have a webhook and we have a webhook now
            webhook_response = await admyral_store.store_workflow_webhook(
                workflow.workflow_id
            )
        else:
            # Case: we had a webhhok and we still have a webhook
            webhook_response = existing_webhook
    elif existing_webhook:
        # Case: we had a webhook but it was removed in the newest push
        await admyral_store.delete_webhook(existing_webhook.webhook_id)

    # construct response
    response = WorkflowPushResponse()
    if webhook_response:
        response.webhook_id = webhook_response.webhook_id
        response.webhook_secret = webhook_response.webhook_secret
    return response


router = APIRouter()


@router.post("/{workflow_name}/push", status_code=status.HTTP_201_CREATED)
async def push_workflow(
    workflow_name: str, request: WorkflowPushRequest
) -> WorkflowPushResponse:
    """
    Push a workflow to the store. If the workflow for the provided workflow id already exists,
    it will be overwritten.

    Args:
        workflow_name: The workflow name.
        workflow: The workflow object.
    """
    return await push_workflow_impl(workflow_name, None, request)


@router.get("/{workflow_name}", status_code=status.HTTP_200_OK)
async def get_workflow(workflow_name: str) -> Optional[Workflow]:
    """
    Get a workflow by its workflow name.

    Args:
        workflow_name: The workflow name.

    Returns:
        The workflow object.
    """
    return await get_admyral_store().get_workflow_by_name(workflow_name)


@router.get("", status_code=status.HTTP_200_OK)
async def list_workflows() -> list[WorkflowMetadata]:
    """
    List all available workflows.

    Returns:
        List of workflow objects.
    """
    return await get_admyral_store().list_workflows()


# TODO: code duplication with webhook_endpoints.py
def exec_task(
    workflow_id: str, payload: dict[str, JsonValue]
) -> Callable[[], Coroutine]:
    async def task() -> None:
        await get_workers_client().execute_workflow(
            workflow_id, MANUAL_TRIGGER_SOURCE_NAME, payload
        )

    return task


@router.post("/{workflow_name}/trigger", status_code=status.HTTP_201_CREATED)
async def trigger_workflow(
    workflow_name: str,
    background_tasks: BackgroundTasks,
    payload: Annotated[dict[str, JsonValue], Body()] = {},
) -> WorkflowTriggerResponse:
    """
    Trigger a workflow.

    Args:
        workflow_name: The workflow name.
        payload: The payload to pass to the workflow.
    """
    workflow = await get_admyral_store().get_workflow_by_name(workflow_name)
    if not workflow:
        raise ValueError(f"Workflow with name {workflow_name} not found.")
    if not workflow.is_active:
        return WorkflowTriggerResponse.inactive()
    background_tasks.add_task(exec_task(workflow.workflow_id, payload))
    return WorkflowTriggerResponse.success()


@router.post("/activate", status_code=status.HTTP_201_CREATED)
async def activate_workflow(body: WorkflowBody) -> bool:
    """
    Activate a workflow.

    Args:
        workflow_name: The workflow name.

    Returns:
        is_active: True if the workflow was successfully activated.
    """
    # if already active, do nothing
    # else we update the state and schedule in temporal

    workflow = await _fetch_workflow(body.workflow_name, body.workflow_id)

    if not workflow.is_active:
        await get_admyral_store().set_workflow_active_state(
            workflow.workflow_id, is_active=True
        )

        schedules = await get_admyral_store().list_schedules_for_workflow(
            workflow.workflow_id
        )
        for schedule in schedules:
            await get_workers_client().schedule_workflow(workflow, schedule)

    return True


@router.post("/deactivate", status_code=status.HTTP_201_CREATED)
async def deactivate_workflow(body: WorkflowBody) -> bool:
    """
    Deactivate a workflow.

    Args:
        workflow_name: The workflow name.

    Returns:
        is_active: False if the workflow was successfully deactivated.
    """
    # if already inactive, do nothing
    # else we update the state and delete the schedule in temporal

    workflow = await _fetch_workflow(body.workflow_name, body.workflow_id)

    if workflow.is_active:
        await get_admyral_store().set_workflow_active_state(
            workflow.workflow_id, is_active=False
        )

        schedules = await get_admyral_store().list_schedules_for_workflow(
            workflow.workflow_id
        )
        for schedule in schedules:
            await get_workers_client().delete_schedule(schedule.schedule_id)

    return False


@router.delete("", status_code=status.HTTP_204_NO_CONTENT)
async def delete_workflow(body: WorkflowBody) -> None:
    """
    Delete a workflow by its workflow name.

    Args:
        workflow_name: The workflow name.
    """
    workflow = await _fetch_workflow(body.workflow_name, body.workflow_id)

    # we must clean up the schedules in temporal if there are any
    schedules = await get_admyral_store().list_schedules_for_workflow(
        workflow.workflow_id
    )
    for schedule in schedules:
        await get_admyral_store().delete_schedule(schedule.schedule_id)
        await get_workers_client().delete_schedule(schedule.schedule_id)

    # now, we can delete the workflow
    await get_admyral_store().remove_workflow(workflow.workflow_id)



import os
from abc import abstractmethod
import json

from admyral.secret.secrets_manager import SecretsManager
from admyral.utils.future_executor import execute_future


class _SecretsAccessImpl:
    @abstractmethod
    def get(self, secret_placeholder: str) -> dict[str, str]: ...

    @abstractmethod
    async def aget(self, secret_placeholder: str) -> dict[str, str]: ...


class SecretsStoreAccessImpl(_SecretsAccessImpl):
    def __init__(
        self,
        user_id: str,
        secret_mappings: dict[str, str],
        secrets_manager: SecretsManager,
    ) -> None:
        """
        Args:
            user_id: The user id of the user who is accessing the secrets.
            secret_mappings: The secret mappings for the action. Defined in the workflow function
            secrets_manager: The secrets manager to use.
        """
        self.user_id = user_id
        self.secret_mappings = secret_mappings
        self.secrets_manager = secrets_manager

    def get(self, secret_placeholder: str) -> dict[str, str]:
        return execute_future(self.aget(secret_placeholder))

    async def aget(self, secret_placeholder: str) -> dict[str, str]:
        secret_name = self.secret_mappings.get(secret_placeholder)
        if not secret_name:
            raise ValueError(
                f"No secret mapped to secret placeholder '{secret_placeholder}'."
            )
        secret = await self.secrets_manager.get(self.user_id, secret_name)
        if not secret:
            raise ValueError(f"Secret '{secret_name}' not found.")
        return secret.secret


class EnvVariableSecretsAccessImpl(_SecretsAccessImpl):
    def get(self, secret_placeholder: str) -> dict[str, str]:
        secret_value = os.environ[secret_placeholder]
        if not secret_value:
            raise ValueError(
                f"Secret placeholder '{secret_placeholder}' not found in environment variables."
            )
        return json.loads(secret_value)

    async def aget(self, secret_placeholder: str) -> dict[str, str]:
        return self.get(secret_placeholder)


class Secrets:
    """
    Used to access secrets in the actions.
    """

    def __init__(self, secrets_access_impl: _SecretsAccessImpl) -> None:
        self.secrets_access_impl = secrets_access_impl

    def get(self, secret_placeholder: str) -> dict[str, str]:
        return self.secrets_access_impl.get(secret_placeholder)

    async def aget(self, secret_placeholder: str) -> dict[str, str]:
        return await self.secrets_access_impl.aget(secret_placeholder)

    @classmethod
    def default(cls) -> "Secrets":
        return Secrets(EnvVariableSecretsAccessImpl())



from admyral.cli.action import (
    action,
    push as action_push,
)
from admyral.cli.server import up, down
from admyral.cli.setup import init
from admyral.cli.workflow import workflow, push as workflow_push, trigger
from admyral.cli.secret import secret, set, delete, list

__all__ = [
    "action",
    "action_push",
    "up",
    "down",
    "init",
    "workflow",
    "workflow_push",
    "trigger",
    "secret",
    "set",
    "delete",
    "list",
]



import inspect
from typing import TYPE_CHECKING, Any, Callable, TypeVar
import ast
from pydantic import BaseModel
import sys

from admyral.models import ActionMetadata
from admyral.utils.json import throw_if_not_allowed_return_type
from admyral.compiler.action_parser import collect_action_arguments

if TYPE_CHECKING:
    F = TypeVar("F", bound=Callable[..., Any])


class ArgumentMetadata(BaseModel):
    display_name: str
    description: str | None = None


def _register_action(action: "Action") -> None:
    from admyral.action_registry import ActionRegistry

    ActionRegistry.register(action.action_type, action)


class Action:
    def __init__(
        self,
        func: "F",
        display_name: str,
        display_namespace: str,
        description: str | None = None,
        secrets_placeholders: list[str] = [],
        requirements: list[str] = [],
    ) -> None:
        self.display_name = display_name
        self.display_namespace = display_namespace
        self.description = description
        self.secrets_placeholders = secrets_placeholders
        self.requirements = requirements

        if self.secrets_placeholders and len(self.secrets_placeholders) != len(
            set(self.secrets_placeholders)
        ):
            raise ValueError("Secret placeholders must be unique.")

        for req in self.requirements:
            if req in sys.stdlib_module_names:
                raise ValueError(
                    f"Standard library module '{req}' should not be added to requirements because they are by default accessible."
                )

        self.func = func
        if self.is_async:
            raise ValueError("Actions cannot be async functions yet.")
        _register_action(self)

        # parse function arguments
        func_def = ast.parse(inspect.getsource(self.func)).body[0]
        self.arguments = collect_action_arguments(func_def)

    @property
    def action_type(self) -> str:
        return self.func.__name__

    @property
    def is_async(self) -> bool:
        return inspect.iscoroutinefunction(self.func)

    def _call(self, **params: dict[str, Any]) -> Any:
        result = self.func(**params)
        throw_if_not_allowed_return_type(result)
        return result

    async def _acall(self, **params: dict[str, Any]) -> Any:
        result = await self.func(**params)
        throw_if_not_allowed_return_type(result)
        return result

    def __call__(
        self,
        *,
        run_after: list[Any] = [],  # compile-time argument
        secrets: dict[str, str] = {},  # compile-time argument
        **params: Any,
    ) -> Any:
        if self.is_async:
            return self._acall(**params)
        return self._call(**params)

    # TODO: currently unused - remmove?
    def get_parameter_names(self) -> list[str]:
        func_def = ast.parse(inspect.getsource(self.func)).body[0]
        args = [arg.arg for arg in func_def.args.args]
        kwonly_args = [arg.arg for arg in func_def.args.kwonlyargs]
        if len(kwonly_args) > 1 or (
            len(kwonly_args) == 1 and kwonly_args[0] != "run_after"
        ):
            raise RuntimeError(
                "Only 'run_after' keyword-only argument is supported for actions"
            )
        return args

    def to_metadata(self) -> ActionMetadata:
        return ActionMetadata(
            action_type=self.action_type,
            display_name=self.display_name,
            display_namespace=self.display_namespace,
            description=self.description,
            secrets_placeholders=self.secrets_placeholders,
            arguments=self.arguments,
        )


def action(
    _func: "F" = None,
    *,
    display_name: str,
    display_namespace: str,
    description: str | None = None,
    secrets_placeholders: list[str] = [],
    requirements: list[str] = [],
) -> Action:
    """Decorator to create a workflow action."""

    def inner(func: "F") -> Action:
        # wrap the action with the temporal activity decorator
        action = Action(
            func,
            display_name=display_name,
            display_namespace=display_namespace,
            description=description,
            secrets_placeholders=secrets_placeholders,
            requirements=requirements,
        )
        action.__doc__ = func.__doc__
        return action

    return inner if _func is None else inner(_func)



import hashlib
import hmac
import os
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from base64 import urlsafe_b64encode, urlsafe_b64decode

from admyral.config.config import WEBHOOK_SIGNING_SECRET, SECRETS_ENCRYPTION_KEY


def _generate_hs256(secret: bytes, data: str) -> str:
    return hmac.new(secret, data.encode(), hashlib.sha256).hexdigest()


def generate_hs256(data: str) -> str:
    return _generate_hs256(WEBHOOK_SIGNING_SECRET, data)


def encrypt_aes256_gcm(secret_key: bytes, plaintext: str) -> str:
    # Generate a random 96-bit IV (Initialization Vector)
    iv = os.urandom(12)

    # Create AES-GCM cipher object
    aes_gcm = AESGCM(secret_key)

    # Encrypt the plaintext and get the associated ciphertext
    ciphertext = aes_gcm.encrypt(iv, plaintext.encode(), None)

    # Concatenate IV and ciphertext and encode to base64
    iv_ciphertext = iv + ciphertext
    iv_ciphertext_b64 = urlsafe_b64encode(iv_ciphertext)

    return iv_ciphertext_b64.decode()


def encrypt_secret(plaintext: str) -> str:
    return encrypt_aes256_gcm(SECRETS_ENCRYPTION_KEY, plaintext)


def decrypt_aes256_gcm(secret_key: bytes, iv_ciphertext_b64: str) -> str:
    # Decode the base64 encoded IV and ciphertext
    iv_ciphertext = urlsafe_b64decode(iv_ciphertext_b64)

    # Extract the IV (first 12 bytes) and the ciphertext
    iv = iv_ciphertext[:12]
    ciphertext = iv_ciphertext[12:]

    # Create AES-GCM cipher object
    aes_gcm = AESGCM(secret_key)

    # Decrypt the ciphertext
    plaintext = aes_gcm.decrypt(iv, ciphertext, None)

    return plaintext.decode()


def decrypt_secret(ciphertext: str) -> str:
    return decrypt_aes256_gcm(SECRETS_ENCRYPTION_KEY, ciphertext)



import ast
import astor
import sys

from admyral.models import Argument, PythonAction
from admyral.typings import JsonValue


def parse_action(module_str: str, action_name: str) -> PythonAction:
    """
    Parse an action function from a module string and return a PythonAction model.

    Args:
        module_str: The module string containing the action function.
        action_name: The name of the action function to parse.

    Returns:
        PythonAction: The parsed action function as a PythonAction model.
    """
    module = ast.parse(module_str)
    module_body = module.body

    # find the action function AST node
    action_node = None
    for node in module_body:
        if (
            isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
            and node.name == action_name
        ):
            action_node = node
            break

    if not action_node:
        raise ValueError(f"Action function '{action_name}' not found in module.")

    action_args = collect_action_arguments(action_node)

    # remove the action decorator from the function
    decorator_list = action_node.decorator_list
    action_node.decorator_list = []

    action_model_params = {
        "action_type": action_name,
        "import_statements": "",
        "code": astor.to_source(action_node),
        "arguments": action_args,
    }

    # extract the action decorator arguments
    if (
        len(decorator_list) != 1
        or (
            isinstance(decorator_list[0], ast.Call)
            and decorator_list[0].func.id != "action"
        )
        or (
            isinstance(decorator_list[0], ast.Name) and decorator_list[0].id != "action"
        )
    ):
        raise ValueError("Action function should only have the @action decorator.")
    action_decorator = decorator_list[0]

    decorator_args = {}

    if isinstance(action_decorator, ast.Call):
        for arg in action_decorator.keywords:
            decorator_args[arg.arg] = arg.value

        for string_arg in ["display_name", "display_namespace", "description"]:
            if value := decorator_args.get(string_arg):
                action_model_params[string_arg] = _parse_constant_string_value(
                    value, string_arg
                )

        for list_arg in ["secrets_placeholders", "requirements"]:
            if value := decorator_args.get(list_arg):
                action_model_params[list_arg] = _parse_list_of_strings(value, list_arg)

    # additionally, we must collect the used import statements and their aliases
    # as well as their imported functions
    # note: we automatically collect all imports from admyral or the standard library as well
    requirements_set = (
        set([req.split("==")[0] for req in action_model_params.get("requirements", [])])
        | sys.stdlib_module_names
    )
    requirements_set.add("admyral")
    requirements_set.add("typing")
    action_model_params["import_statements"] = "".join(
        _collect_imports(module_body, requirements_set)
    )

    # filter out stdlib modules from the requirements
    if requirements := action_model_params.get("requirements"):
        action_model_params["requirements"] = [
            req for req in requirements if req not in sys.stdlib_module_names
        ]

    return PythonAction.model_validate(action_model_params)


def _parse_constant_string_value(node: ast.expr, parameter_name: str) -> str:
    if not isinstance(node, ast.Constant) or not isinstance(node.value, str):
        raise ValueError(f"Decorator argument for {parameter_name} must be a string.")
    return node.value


def _parse_list_of_strings(node: ast.expr, parameter_name: str) -> list[str]:
    if not isinstance(node, ast.List):
        raise ValueError(
            f"Decorator argument for {parameter_name} must be a list of strings."
        )
    values = []
    for elt in node.elts:
        if not isinstance(elt, ast.Constant) or not isinstance(elt.value, str):
            raise ValueError(
                f"Decorator argument for {parameter_name} must be a list of strings."
            )
        values.append(elt.value)
    return values


def _collect_imports(
    module_body: list[ast.stmt], requirements_set: set[str]
) -> list[str]:
    imports = list(
        filter(lambda node: isinstance(node, (ast.Import, ast.ImportFrom)), module_body)
    )

    collected_imports = []
    for imp in imports:
        if isinstance(imp, ast.Import):
            for alias in imp.names:
                package_name = alias.name.split(".")[0]
                if package_name in requirements_set:
                    collected_imports.append(
                        astor.code_gen.to_source(ast.Import(names=[alias]))
                    )

        if isinstance(imp, ast.ImportFrom):
            package_name = imp.module.split(".")[0]
            if package_name in requirements_set:
                collected_imports.append(astor.code_gen.to_source(imp))

    return collected_imports


class IsOptionalTypeVisitor(ast.NodeVisitor):
    def __init__(self):
        self.is_optional = False

    def visit_Constant(self, node: ast.Constant):
        if node.value is None:
            self.is_optional = True

    def visit_Subscript(self, node: ast.Subscript):
        if node.value.id == "Optional":
            self.is_optional = True


def is_optional_type(arg_type_ast: ast.arg) -> bool:
    """
    Check whether an argument is optional. An argument is optional if its type is
    explicitly annotated with None or Optional.

    Args:
        arg: The AST node of the argument.

    Returns:
        bool: True if the argument is optional, False otherwise.
    """
    is_optional_visitor = IsOptionalTypeVisitor()
    is_optional_visitor.visit(arg_type_ast)
    return is_optional_visitor.is_optional


def _parse_type_annotation(arg: ast.arg) -> tuple[str, bool, dict[str, str]]:
    """
    Parse the type annotation of an argument, i.e.:

    Annotated[
        <type>,
        ArgumentMetadata(...)
    ]

    Args:
        arg: The AST node of the argument.

    Returns:
        tuple[str, bool, ArgumentMetadata]: The argument type, whether the argument is optional, and the argument metadata.
    """
    from admyral.action import ArgumentMetadata

    # extract argument type and metadata
    subscript = arg.annotation
    if (
        not isinstance(subscript, ast.Subscript)
        or subscript.value.id != "Annotated"
        or not isinstance(subscript.slice, ast.Tuple)
        or len(subscript.slice.elts) != 2
    ):
        raise ValueError(
            "Arguments must be annotated using Annotated[<type>, ArgumentMetadata(...)]."
        )

    arg_type_ast = subscript.slice.elts[0]
    argument_metadata_ast = subscript.slice.elts[1]

    # get type and check whether it is optional
    arg_type = astor.to_source(arg_type_ast).strip()  # TODO: improve this
    is_optional = is_optional_type(arg_type_ast)

    # get ArgumentMetadata info
    if (
        not isinstance(argument_metadata_ast, ast.Call)
        or argument_metadata_ast.func.id != "ArgumentMetadata"
    ):
        raise ValueError(
            "Argument metadata must be provided using ArgumentMetadata(...)."
        )

    arg_metadata_parmas = {}
    for keyword in argument_metadata_ast.keywords:
        if not isinstance(keyword.value, ast.Constant) or not isinstance(
            keyword.value.value, str
        ):
            raise ValueError("Argument metadata parameters must be a string.")
        if keyword.arg in arg_metadata_parmas:
            raise ValueError(
                f"Found duplicate ArgumentMetadata parameter: {keyword.arg}. ArgumentMetadata parameters must be unique."
            )
        arg_metadata_parmas[keyword.arg] = keyword.value.value
    arg_metadata = ArgumentMetadata.model_validate(arg_metadata_parmas)

    return arg_type, is_optional, arg_metadata.model_dump()


def _parse_value(value_ast: ast.expr) -> JsonValue:
    """
    Transform AST value expression into a JSON-compatible value.

    Args:
        value_ast: The AST node of the value expression.

    Returns:
        JsonValue: The JSON-compatible value.
    """
    if isinstance(value_ast, ast.Constant):
        return value_ast.value

    if isinstance(value_ast, ast.List):
        return [_parse_value(elt) for elt in value_ast.elts]

    if isinstance(value_ast, ast.Dict):
        return {
            key.value: _parse_value(value)
            for key, value in zip(value_ast.keys, value_ast.values)
        }

    expr = astor.to_source(value_ast)
    raise ValueError(
        f'Expression "{expr}" as default value not supported. Default value must be JSON-compatible: constant, list, or dict.'
    )


def collect_action_arguments(action_node: ast.FunctionDef) -> list[Argument]:
    """
    Collect the arguments of an action function.

    Args:
        action_node: The AST node of the action function.

    Returns:
        list[Argument]: The collected arguments as Argument models.
    """
    action_def_args = action_node.args

    # Note: the last #defaults elements of args in the args list, are the corresponding default values
    num_of_args_without_default_value = len(action_def_args.args) - len(
        action_def_args.defaults
    )

    arguments = []
    for idx, arg in enumerate(action_node.args.args):
        # check for default value
        default_value = None
        if idx >= num_of_args_without_default_value:
            default_value_idx = idx - num_of_args_without_default_value
            default_value_ast = action_def_args.defaults[default_value_idx]
            default_value = _parse_value(default_value_ast)

        # parse type annotation
        arg_type, is_optional, arg_metadata = _parse_type_annotation(arg)

        arguments.append(
            Argument.model_validate(
                {
                    "arg_name": arg.arg,
                    "display_name": arg_metadata["display_name"],
                    "description": arg_metadata.get("description"),
                    "arg_type": arg_type,
                    "is_optional": is_optional,
                    "default_value": default_value,
                }
            )
        )

    for arg, default_value_ast in zip(
        action_def_args.kwonlyargs, action_def_args.kw_defaults
    ):
        default_value = _parse_value(default_value_ast)
        arg_type, is_optional, arg_metadata = _parse_type_annotation(arg)
        arguments.append(
            Argument.model_validate(
                {
                    "arg_name": arg.arg,
                    "display_name": arg_metadata["display_name"],
                    "description": arg_metadata.get("description"),
                    "arg_type": arg_type,
                    "is_optional": is_optional,
                    "default_value": default_value,
                }
            )
        )

    return arguments



import asyncio
import os
from typing import Optional, Generator
import time

from admyral.services.worker_service import launch_worker
from admyral.services.api_service import run_api
from admyral.services.temporal_service import run_temporal, start_after_temporal
from admyral.utils.daemon import launch_as_daemon, stop_daemon
from admyral.config.config import (
    get_admyral_daemon_log_file,
    get_admyral_daemon_pid_file,
    get_temporal_daemon_log_file,
    get_temporal_daemon_pid_file,
    GlobalConfig,
)
from admyral.utils.io import file_tail


async def launch_admyral_blocking():
    """
    Starts the following Admyral services:
    - Temporal
    - Admyral Worker
    - Admyral Backend Server

    The method first starts Temporal, then starts the worker and backend server
    after Temporal was successfully launched.
    """
    await asyncio.gather(
        run_temporal(),
        start_after_temporal(
            [launch_worker({"docker": False}), run_api()], GlobalConfig().temporal_host
        ),
    )


def launch_admyral_daemon():
    print("Starting Temporal...")
    launch_as_daemon(
        lambda: asyncio.run(run_temporal()),
        get_temporal_daemon_pid_file(),
        get_temporal_daemon_log_file(),
    )
    print("Started Temporal.")

    def run_worker_and_api_after_temporal():
        asyncio.run(
            start_after_temporal(
                [launch_worker({"docker": False}), run_api()],
                GlobalConfig().temporal_host,
            )
        )

    print("Starting Admyral...")
    launch_as_daemon(
        run_worker_and_api_after_temporal,
        get_admyral_daemon_pid_file(),
        get_admyral_daemon_log_file(),
        os.getcwd(),
    )
    print("Started Admyral.")


def destroy_admyral_daemon():
    # we first stop the worker and the backend server
    print("Stopping Admyral...")
    stop_daemon(get_admyral_daemon_pid_file())
    print("Stopped Admyral.")

    # then we stop Temporal
    print("Stopping Temporal...")
    stop_daemon(get_temporal_daemon_pid_file())
    print("Stopped Temporal.")


def show_logs(
    follow: bool = False, tail: Optional[int] = None
) -> Generator[str, None, None]:
    log_file = get_admyral_daemon_log_file()
    if not os.path.exists(log_file):
        return

    with open(log_file, "r") as f:
        if tail is not None:
            if tail < 0:
                raise ValueError("tail must be a positive integer.")
            # TODO: make file_tail a generator
            available_bytes, tail_lines = file_tail(log_file, tail)
            for line in tail_lines:
                yield line

            if not follow:
                return

            # we need to set the file cursor until where we read the last lines
            f.seek(available_bytes)

        line = ""
        while True:
            if partial_line := f.readline():
                line += partial_line
                if line.endswith("\n"):
                    # we have a full line, so we can now emit it
                    yield line
                    line = ""
            elif follow:
                # wait for new lines to be written to the log file
                time.sleep(1)
            else:
                # no more lines to read
                break


# TODO: show dashboard, API URL
def show_status():
    admyral_pid_file = get_admyral_daemon_pid_file()
    temporal_pid_file = get_temporal_daemon_pid_file()

    if os.path.exists(admyral_pid_file):
        print("Admyral is running.")
    else:
        print("Admyral is not running.")
    if os.path.exists(temporal_pid_file):
        print("Temporal is running.")
    else:
        print("Temporal is not running.")



"""
TODO:

- all logs should have the same format
    - adapt log format for uvicorn: https://github.com/encode/uvicorn/discussions/2027

"""

import logging

from admyral.config.config import LOGGING_LEVEL, LoggingLevel


def get_log_level() -> LoggingLevel:
    if LOGGING_LEVEL not in LoggingLevel.__members__:
        raise ValueError(f"Invalid logging level: {LOGGING_LEVEL}")
    return LoggingLevel[LOGGING_LEVEL]


def get_stream_handler() -> logging.StreamHandler:
    stream_handler = logging.StreamHandler()
    # log_formatter = logging.Formatter("%(asctime)s [%(processName)s: %(process)d] [%(threadName)s: %(thread)d] [%(levelname)s] %(name)s: %(message)s")
    log_formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
    )
    stream_handler.setFormatter(log_formatter)
    return stream_handler


def setup_root_logger() -> None:
    logging.root.setLevel(get_log_level().value)
    logging.root.addHandler(get_stream_handler())


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(get_log_level().value)
    logger.addHandler(get_stream_handler())
    logger.propagate = False
    return logger



from sqlmodel import Field
from sqlalchemy import TEXT, JSON

from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import PythonAction, ActionMetadata
from admyral.typings import JsonValue


class PythonActionSchema(BaseSchema, table=True):
    """
    Schema for Python Actions
    """

    __tablename__ = "python_actions"

    # primary keys
    user_id: str = Field(sa_type=TEXT(), primary_key=True)  # TODO: link to users table
    action_type: str = Field(sa_type=TEXT(), primary_key=True)

    # other fields
    import_statements: str = Field(sa_type=TEXT())
    code: str = Field(sa_type=TEXT())
    display_name: str = Field(sa_type=TEXT())
    display_namespace: str = Field(sa_type=TEXT())
    description: str | None = Field(sa_type=TEXT(), nullable=True)
    secrets_placeholders: str | None = Field(sa_type=TEXT(), nullable=True)
    requirements: str | None = Field(sa_type=TEXT(), nullable=True)
    arguments: list[dict[str, JsonValue]] = Field(sa_type=JSON())

    def to_model(self, include_resources: bool = False) -> PythonAction:
        return PythonAction.model_validate(
            {
                "action_type": self.action_type,
                "import_statements": self.import_statements,
                "code": self.code,
                "display_name": self.display_name,
                "display_namespace": self.display_namespace,
                "description": self.description,
                "secrets_placeholders": self.secrets_placeholders.split(";")
                if self.secrets_placeholders
                else [],
                "requirements": self.requirements.split(";")
                if self.requirements
                else [],
                "arguments": self.arguments,
            }
        )

    def to_metadata(self) -> ActionMetadata:
        return ActionMetadata.model_validate(
            {
                "action_type": self.action_type,
                "display_name": self.display_name,
                "display_namespace": self.display_namespace,
                "description": self.description,
                "secrets_placeholders": self.secrets_placeholders.split(";")
                if self.secrets_placeholders
                else [],
                "arguments": self.arguments,
            }
        )
